#include"cmplr_literal.hpp"
#include"cmplr_function.hpp"
#include"expression_node.hpp"
#include"cmplr_precontext.hpp"
#include"cmplr_context.hpp"




Literal::
Literal():
kind(LiteralKind::null)
{
}


Literal::Literal( int32_t             i): kind(LiteralKind::integer      ){data.i     = i;}
Literal::Literal(char16_t             c): kind(LiteralKind::character    ){data.i     = c;}
Literal::Literal(nullptr_t          nul): kind(LiteralKind::null_pointer ){}
Literal::Literal(bool                 b): kind(LiteralKind::boolean      ){data.i     = b;}
Literal::Literal(std::u16string*      s): kind(LiteralKind::string       ){data.s     = s;}
Literal::Literal(std::string*        id): kind(LiteralKind::identifier   ){data.id    = id;}
Literal::Literal(expression::Node*   nd): kind(LiteralKind::expression   ){data.nd    = nd;}
Literal::Literal(Literal*           lit): kind(LiteralKind::literal      ){data.lit   = lit;}
Literal::Literal(Array*             arr): kind(LiteralKind::array        ){data.arr   = arr;}
Literal::Literal(Function*           fn): kind(LiteralKind::function     ){data.fn    = fn;}
Literal::Literal(ArgumentList*     args): kind(LiteralKind::argument_list){data.args  = args;}
Literal::Literal(Subscript*       subsc): kind(LiteralKind::subscript    ){data.subsc = subsc;}


Literal::
Literal(const mkf::Node&  src, PreContext&  prectx):
kind(LiteralKind::null)
{
  read(src,prectx);
}


Literal::
Literal(const Literal&  rhs) noexcept:
kind(LiteralKind::null)
{
  *this = rhs;
}


Literal::
Literal(Literal&&  rhs) noexcept:
kind(LiteralKind::null)
{
  *this = std::move(rhs);
}


Literal::
~Literal()
{
  clear();
}




Literal&
Literal::
operator=(const Literal&   rhs) noexcept
{
  clear();

  kind = rhs.kind;

    switch(kind)
    {
  case(LiteralKind::null_pointer):
      break;
  case(LiteralKind::string):
      data.s = new std::u16string(*rhs.data.s);
      break;
  case(LiteralKind::expression):
      data.nd = new expression::Node(*rhs.data.nd);
      break;
  case(LiteralKind::literal):
      data.lit = new Literal(*rhs.data.lit);
      break;
  case(LiteralKind::subscript):
      data.subsc = new Subscript(*rhs.data.subsc);
      break;
  case(LiteralKind::array):
      data.arr = new Array(*rhs.data.arr);
      break;
  case(LiteralKind::argument_list):
      data.args = new ArgumentList(*rhs.data.args);
      break;
  case(LiteralKind::identifier):
      data.id = new std::string(*rhs.data.id);
      break;
  case(LiteralKind::function):
      data.fn = new Function(*rhs.data.fn);
      break;
  case(LiteralKind::integer):
  case(LiteralKind::character):
  case(LiteralKind::boolean):
      data.i = rhs.data.i;
      break;
    }


  return *this;
}


Literal&
Literal::
operator=(Literal&&  rhs) noexcept
{
  clear();

  kind = rhs.kind                    ;
         rhs.kind = LiteralKind::null;

  data = rhs.data;


  return *this;
}




void
Literal::
clear()
{
    switch(kind)
    {
  case(LiteralKind::identifier):
      delete data.id;
      break;
  case(LiteralKind::string):
      delete data.s;
      break;
  case(LiteralKind::expression):
      delete data.nd;
      break;
  case(LiteralKind::subscript):
      delete data.subsc;
      break;
  case(LiteralKind::literal):
      delete data.lit;
      break;
  case(LiteralKind::array):
      delete data.arr;
      break;
  case(LiteralKind::argument_list):
      delete data.args;
      break;
  case(LiteralKind::function):
      delete data.fn;
      break;
  case(LiteralKind::integer):
  case(LiteralKind::character):
  case(LiteralKind::boolean):
      break;
    }


  kind = LiteralKind::null;
}




Constant
Literal::
make_constant(PreContext&  prectx) const
{
    switch(kind)
    {
  case(LiteralKind::identifier):
      {
        auto  decl = prectx.find_declaration(*data.id);

          if(!decl)
          {
            printf("%sが指すオブジェクトが見つかりません\n",data.id->data());

            throw;
          }


        return decl->literal.make_constant(prectx);
      }
      break;
  case(LiteralKind::string):
      return Constant(*data.s);
      break;
  case(LiteralKind::expression):
      return data.nd->make_constant(prectx);
      break;
  case(LiteralKind::array):
    {
      std::vector<Constant>  ls;

        for(auto&  lit: *data.arr)
        {
          ls.emplace_back(lit.make_constant(prectx));
        }


      return Constant(ls);
    }
     break;
  case(LiteralKind::argument_list):
      break;
  case(LiteralKind::subscript):
      break;
  case(LiteralKind::function):
      break;
  case(LiteralKind::integer):
      return Constant(data.i);
      break;
  case(LiteralKind::character):
      return Constant(data.i);
      break;
  case(LiteralKind::null_pointer):
      return Constant(0);
      break;
    }


  throw Constant::Error();
}




void
Literal::
compile_data(const Type&  type, Context&  ctx) const
{
    switch(kind)
    {
  case(LiteralKind::function):
      data.fn->compile_definition(ctx);
      break;
/*
  case(LiteralKind::array):
      break;
  case(LiteralKind::expression):
      compile_value(ctx);
      break;
  case(LiteralKind::literal):
      compile_value(ctx);
      break;
  case(LiteralKind::integer):
      compile_value(ctx);
      break;
  case(LiteralKind::character):
      compile_value(ctx);
      break;
*/    }
}




Type
Literal::
compile_value(Context&  ctx) const
{
    switch(kind)
    {
  case(LiteralKind::function):
    return data.fn->compile_call(ctx);
    break;
  case(LiteralKind::identifier):
    {
      auto  res = ctx.find_declaration(*data.id);

        if(!res)
        {
          printf("関数%s内において、識別子%sが指すオブジェクトが見つかりません\n",
                 ctx.function->name.data(),data.id->data());

          throw;
        }


      return res->compile_use(ctx);
    } break;
  case(LiteralKind::expression):
      return data.nd->compile_value(ctx);
      break;
  case(LiteralKind::literal):
      return data.lit->compile_value(ctx);
      break;
  case(LiteralKind::array):
      break;
  case(LiteralKind::subscript):
      break;
  case(LiteralKind::argument_list):
    {
      auto&  ls = *data.arr;

        if(ls.size())
        {
          ctx.push("  //------------//\n");
          ctx.push("  //引数積み始め//\n");
          ctx.push("  //------------//\n");

/*
            for(auto  it = data.arr->crbegin();  it != data.arr->crend();  ++it)
            {
              auto  t = it->compile(ctx);

                if(t == TypeKind::reference)
                {
                  t = t.compile_dereference(ctx);
                }
            }
*/

          ctx.push("  //--------------//\n");
          ctx.push("  //引数積み終わり//\n");
          ctx.push("  //--------------//\n");
        }


      return Type(data.args);
    } break;
  case(LiteralKind::character):
           if(data.i <= 0x00FF){ctx.push("  psh8u   %d;//即値\n",data.i);}
      else if(data.i <= 0xFFFF){ctx.push("  psh16u  %d;//即値\n",data.i);}
      else                     {ctx.push("  psh32   %d;//即値\n",data.i);}

      return ConstCharType();
      break;
  case(LiteralKind::integer):
           if(data.i <= 0x00FF){ctx.push("  psh8u   %d;//即値\n",data.i);}
      else if(data.i <= 0xFFFF){ctx.push("  psh16u  %d;//即値\n",data.i);}
      else                     {ctx.push("  psh32   %d;//即値\n",data.i);}

      return ConstI32Type();
      break;
    }


  return Type();
}




void
Literal::
print(FILE*  f) const
{
    switch(kind)
    {
  case(LiteralKind::character):
      fprintf(f,"\'%s\'",pp::UTF8Chunk(data.i).codes);
      break;
  case(LiteralKind::expression):
      data.nd->print(f);
      break;
  case(LiteralKind::identifier):
      fprintf(f,"%s",data.id->data());
      break;
  case(LiteralKind::string):
      fprintf(f,"\"");

        for(auto  c: *data.s)
        {
          fprintf(f,"%s",pp::UTF8Chunk(c).codes);
        }


      fprintf(f,"\"");
      break;
  case(LiteralKind::subscript):
      fprintf(f,"[");
      data.subsc->print(f);
      fprintf(f,"]");
      break;
  case(LiteralKind::argument_list):
    {
      fprintf(f,"<");

      auto   it = data.args->cbegin();
      auto  end = data.args->cend();

        if(it != end)
        {
          it++->print(f);

            while(it != end)
            {
              fprintf(f,",");

              it++->print(f);
            }
        }

      fprintf(f,">");
    } break;
  case(LiteralKind::integer):
      fprintf(f,"%d",data.i);
      break;
  case(LiteralKind::null_pointer):
      fprintf(f,"nullptr");
      break;
  case(LiteralKind::boolean):
      fprintf(f,"%s",data.i? "true":"false");
      break;
  case(LiteralKind::literal):
      data.lit->print(f);
      break;
  case(LiteralKind::array):
      fprintf(f,"{");

        for(auto&  e: *data.arr)
        {
          e.print(f);

          fprintf(f,",");
        }


      fprintf(f,"}");
      break;
  case(LiteralKind::function):
      data.fn->print(f);
      break;
    }
}




void
Literal::
read(const mkf::Node&  src, PreContext&  prectx)
{
  mkf::Cursor  cur(src);

    while(!cur.test_ended())
    {
      auto&  nd = cur.get();

        if(nd == "identifier")
        {
          auto  s = new std::string;

          nd.collect_characters(*s);

          *this = Literal(s);
        }

      else
        if(nd == "integer_literal")
        {
          *this = Literal(static_cast<int32_t>(read_integer_literal(nd)));
        }

      else
        if(nd == "character_literal")
        {
          *this = Literal(read_character_literal(nd));
        }

      else
        if(nd == "expression")
        {
          *this = Literal(new expression::Node(nd,prectx));
        }

        if(nd == "string_literal")
        {
          *this = Literal(new std::u16string(read_string_literal(nd)));
        }

      else
        if(nd == "nullptr")
        {
          *this = Literal(nullptr);
        }

      else
        if(nd == "true")
        {
          *this = Literal(true);
        }

      else
        if(nd == "false")
        {
          *this = Literal(false);
        }

      else
        if(nd == "array_literal")
        {
          *this = Literal(new Array(read_list(nd,prectx)));
        }


      cur.advance();
    }
}


Array
Literal::
read_list(const mkf::Node&  src, PreContext&  prectx)
{
  Array  arr;

  mkf::Cursor  cur(src);

    while(!cur.test_ended())
    {
      auto&  nd = cur.get();

        if(nd == "literal_object")
        {
          arr.emplace_back(nd,prectx);
        }

      else
        if(nd == "expression")
        {
          arr.emplace_back(new expression::Node(nd,prectx));
        }


      cur.advance();
    }


  return std::move(arr);
}




